@@ -1,6 +1,6 @@
 /*
  * conn.c
- * $Id: conn.c,v 1.49 2002/04/02 03:49:13 ttate Exp $
+ * $Id: conn.c,v 1.11 2003/12/06 17:33:39 ttate Exp $
  */
 
 #include "ruby.h"
@@ -32,29 +32,36 @@ rb_ldap_conn_new(VALUE klass, LDAP *cldap)
   VALUE conn;
   RB_LDAP_DATA *ldapdata;
 
-  if( cldap != NULL ){
-    conn = Data_Make_Struct(klass, RB_LDAP_DATA,
-			    rb_ldap_conn_mark,rb_ldap_conn_free,ldapdata);
-    ldapdata->ldap = cldap;
-    ldapdata->err  = 0;
-    ldapdata->bind = 0;
-  }
-  else{
-    rb_raise(rb_eLDAP_ResultError,"can't initialize a ldap session");
-  };
+  conn = Data_Make_Struct(klass, RB_LDAP_DATA,
+			  rb_ldap_conn_mark,rb_ldap_conn_free,ldapdata);
+  ldapdata->ldap = cldap;
+  ldapdata->err  = 0;
+  ldapdata->bind = 0;
 
   return conn;
 };
 
+
 VALUE
-rb_ldap_conn_s_new(int argc, VALUE argv[], VALUE klass)
+rb_ldap_conn_s_allocate(VALUE klass)
+{
+  return rb_ldap_conn_new(klass,(LDAP*)0);
+}
+
+VALUE
+rb_ldap_conn_initialize(int argc, VALUE argv[], VALUE self)
 {
   LDAP *cldap;
   char *chost;
   int cport;
+  RB_LDAP_DATA *ldapdata;
 
   VALUE arg1, arg2;
-  VALUE conn;
+
+  Data_Get_Struct(self, RB_LDAP_DATA, ldapdata);
+  if( ldapdata->ldap ){
+    return Qnil;
+  }
 
   switch( rb_scan_args(argc,argv,"02",&arg1,&arg2) ){
   case 0:
@@ -63,22 +70,23 @@ rb_ldap_conn_s_new(int argc, VALUE argv[], VALUE klass)
     cport = LDAP_PORT;
     break;
   case 1:
-    chost = STR2CSTR(arg1);
+    chost = StringValueCStr(arg1);
     cport = LDAP_PORT;
     break;
   case 2:
-    chost = STR2CSTR(arg1);
+    chost = StringValueCStr(arg1);
     cport = NUM2INT(arg2);
     break;
   default:
     rb_bug("rb_ldap_conn_new");
   };
-
+  
   cldap = ldap_init(chost,cport);
-  conn = rb_ldap_conn_new(klass,cldap);
+  if( !cldap )
+    rb_raise(rb_eLDAP_ResultError,"can't initialize a ldap session");
+  ldapdata->ldap = cldap;
 
-  rb_obj_call_init(conn,argc,argv);
-  return conn;
+  return Qnil;
 };
 
 VALUE
@@ -98,11 +106,11 @@ rb_ldap_conn_s_open(int argc, VALUE argv[], VALUE klass)
     cport = LDAP_PORT;
     break;
   case 1:
-    chost = STR2CSTR(arg1);
+    chost = StringValueCStr(arg1);
     cport = LDAP_PORT;
     break;
   case 2:
-    chost = STR2CSTR(arg1);
+    chost = StringValueCStr(arg1);
     cport = NUM2INT(arg2);
     break;
   default:
@@ -110,19 +118,14 @@ rb_ldap_conn_s_open(int argc, VALUE argv[], VALUE klass)
   };
 
   cldap = ldap_open(chost,cport);
+  if( !cldap )
+    rb_raise(rb_eLDAP_ResultError,"can't open a ldap session");
   conn = rb_ldap_conn_new(klass, cldap);
 
-  rb_obj_call_init(conn,argc,argv);
 
   return conn;
 };
 
-static VALUE
-rb_ldap_conn_initialize(int argc, VALUE argv[], VALUE self)
-{
-  return Qnil;
-};
-
 VALUE
 rb_ldap_conn_start_tls_s(int argc, VALUE argv[], VALUE self)
 {
@@ -166,17 +169,32 @@ rb_ldap_conn_simple_bind_s(int argc, VALUE argv[], VALUE self)
 
   GET_LDAP_DATA(self,ldapdata);
   if( ldapdata->bind ){
-    raise(rb_eLDAP_Error, "already binded.");
+    rb_raise(rb_eLDAP_Error, "already binded.");
   };
   switch( rb_scan_args(argc, argv, "02", &arg1, &arg2) ){
   case 0:
     break;
   case 1:
-    dn = STR2CSTR(arg1);
+    if( arg1 == Qnil ){
+      dn = NULL;
+    }
+    else{
+      dn = StringValueCStr(arg1);
+    }
     break;
   case 2:
-    dn = STR2CSTR(arg1);
-    passwd = STR2CSTR(arg2);
+    if( arg1 == Qnil ){
+      dn = NULL;
+    }
+    else{
+      dn = StringValueCStr(arg1);
+    }
+    if( arg2 == Qnil ){
+      passwd = NULL;
+    }
+    else{
+      passwd = StringValueCStr(arg2);
+    }
     break;
   default:
     rb_bug("rb_ldap_conn_simple_bind_s");
@@ -207,21 +225,21 @@ rb_ldap_conn_bind_s(int argc, VALUE argv[], VALUE self)
 
   GET_LDAP_DATA(self,ldapdata);
   if( ldapdata->bind ){
-    raise(rb_eLDAP_Error, "already binded.");
+    rb_raise(rb_eLDAP_Error, "already binded.");
   };
   switch( rb_scan_args(argc, argv, "03", &arg1, &arg2, &arg3) ){
   case 0:
     break;
   case 1:
-    dn = STR2CSTR(arg1);
+    dn = StringValueCStr(arg1);
     break;
   case 2:
-    dn = STR2CSTR(arg1);
-    passwd = STR2CSTR(arg2);
+    dn = StringValueCStr(arg1);
+    passwd = StringValueCStr(arg2);
     break;
   case 3:
-    dn = STR2CSTR(arg1);
-    passwd = STR2CSTR(arg2);
+    dn = StringValueCStr(arg1);
+    passwd = StringValueCStr(arg2);
     method = NUM2INT(arg3);
     break;
   default:
@@ -272,10 +290,12 @@ rb_ldap_conn_set_option(VALUE self, VALUE opt, VALUE data)
   copt = NUM2INT(opt);
 
   switch( copt ){
+  case LDAP_OPT_REFERRALS:
+     optdata = (void*)NUM2INT(data);
+     break;
   case LDAP_OPT_DEREF:
   case LDAP_OPT_SIZELIMIT:
   case LDAP_OPT_TIMELIMIT:
-  case LDAP_OPT_REFERRALS:
   case LDAP_OPT_RESTART:
   case LDAP_OPT_PROTOCOL_VERSION:
   case LDAP_OPT_ERROR_NUMBER:
@@ -322,18 +342,19 @@ rb_ldap_conn_set_option(VALUE self, VALUE opt, VALUE data)
   case LDAP_OPT_X_TLS_RANDOM_FILE:
 #endif
 #endif
-    optdata = STR2CSTR(data);
+    optdata = StringValueCStr(data);
     break;
 #ifdef LDAP_OPT_API_INFO
   case LDAP_OPT_API_INFO:
-    optdata = (void*)rb_ldap_get_apiinfo(data);
+    rb_raise(rb_eLDAP_Error, "option is READ-ONLY");
+    /* optdata = (void*)rb_ldap_get_apiinfo(data); */
     break;
 #endif
   default:
     rb_notimplement();
   }
   ldapdata->err = ldap_set_option(ldapdata->ldap, copt, optdata);
-  Check_LDAP_Result(ldapdata->err);
+  Check_LDAP_OPT_Result(ldapdata->err);
 
   return self;
 #else
@@ -457,7 +478,7 @@ rb_ldap_conn_perror(VALUE self, VALUE msg)
   char *str;
 
   GET_LDAP_DATA(self,ldapdata);
-  cmsg = STR2CSTR(msg);
+  cmsg = StringValueCStr(msg);
 #if defined(HAVE_LDAP_PERROR) && (! defined(USE_NETSCAPE_SDK))
   ldap_perror(ldapdata->ldap, cmsg);
 #else
@@ -570,130 +591,49 @@ rb_ldap_conn_search_i(int argc, VALUE argv[], VALUE self,
 
   GET_LDAP_DATA(self, (*ldapdata));
   cldap = (*ldapdata)->ldap;
+
+  cattrs = NULL;
+  cattrsonly = 0;
+  tv.tv_sec = 0;
+  tv.tv_usec = 0;
+  sort_attr = NULL;
   rb_ldap_sort_obj = Qnil;
 
   switch( rb_scan_args(argc,argv,"36",
 		       &base,&scope,&filter,&attrs,&attrsonly,&sec,&usec,&s_attr,&s_proc) ){
   case 9:
     rb_ldap_sort_obj = s_proc; /* Ruby's GC never start in a C function */
-    sort_attr = STR2CSTR(s_attr);
-    cbase = STR2CSTR(base);
-    cscope = NUM2INT(scope);
-    cfilter = STR2CSTR(filter);
-    cattrsonly = (attrsonly==Qtrue) ? 1 : 0;
-    if( TYPE(attrs) == T_ARRAY ){
-      cattrs = ALLOCA_N(char*,(RARRAY(attrs)->len + 1));
-      for( i=0; i < RARRAY(attrs)->len; i++ ){
-	cattrs[i] = STR2CSTR(RARRAY(attrs)->ptr[i]);
-      };
-      cattrs[RARRAY(attrs)->len] = NULL;
-    }
-    else{
-      cattrs = NULL;
-    };
-    tv.tv_sec = NUM2INT(sec);
-    tv.tv_usec = NUM2INT(usec);
-    break;
   case 8:
-    rb_ldap_sort_obj = Qtrue;
-    sort_attr = STR2CSTR(s_attr);
-    cbase = STR2CSTR(base);
-    cscope = NUM2INT(scope);
-    cfilter = STR2CSTR(filter);
-    cattrsonly = (attrsonly==Qtrue) ? 1 : 0;
-    if( TYPE(attrs) == T_ARRAY ){
-      cattrs = ALLOCA_N(char*,(RARRAY(attrs)->len + 1));
-      for( i=0; i < RARRAY(attrs)->len; i++ ){
-	cattrs[i] = STR2CSTR(RARRAY(attrs)->ptr[i]);
-      };
-      cattrs[RARRAY(attrs)->len] = NULL;
+    if( rb_ldap_sort_obj == Qnil ){
+      rb_ldap_sort_obj = Qtrue;
     }
-    else{
-      cattrs = NULL;
-    };
-    tv.tv_sec = NUM2INT(sec);
-    tv.tv_usec = NUM2INT(usec);
-    break;
+    sort_attr = StringValueCStr(s_attr);
   case 7:
-    cbase = STR2CSTR(base);
-    cscope = NUM2INT(scope);
-    cfilter = STR2CSTR(filter);
-    cattrsonly = (attrsonly==Qtrue) ? 1 : 0;
-    if( TYPE(attrs) == T_ARRAY ){
-      cattrs = ALLOCA_N(char*,(RARRAY(attrs)->len + 1));
-      for( i=0; i < RARRAY(attrs)->len; i++ ){
-	cattrs[i] = STR2CSTR(RARRAY(attrs)->ptr[i]);
-      };
-      cattrs[RARRAY(attrs)->len] = NULL;
-    }
-    else{
-      cattrs = NULL;
-    };
-    tv.tv_sec = NUM2INT(sec);
     tv.tv_usec = NUM2INT(usec);
-    break;
   case 6:
-    cbase = STR2CSTR(base);
-    cscope = NUM2INT(scope);
-    cfilter = STR2CSTR(filter);
-    cattrsonly = (attrsonly==Qtrue) ? 1 : 0;
-    if( TYPE(attrs) == T_ARRAY ){
-      cattrs = ALLOCA_N(char*,(RARRAY(attrs)->len + 1));
-      for( i=0; i < RARRAY(attrs)->len; i++ ){
-	cattrs[i] = STR2CSTR(RARRAY(attrs)->ptr[i]);
-      };
-      cattrs[RARRAY(attrs)->len] = NULL;
-    }
-    else{
-      cattrs = NULL;
-    };
     tv.tv_sec = NUM2INT(sec);
-    tv.tv_usec = 0;
-    break;
   case 5:
-    cbase = STR2CSTR(base);
-    cscope = NUM2INT(scope);
-    cfilter = STR2CSTR(filter);
     cattrsonly = (attrsonly==Qtrue) ? 1 : 0;
-    if( TYPE(attrs) == T_ARRAY ){
-      cattrs = ALLOCA_N(char*,(RARRAY(attrs)->len + 1));
-      for( i=0; i < RARRAY(attrs)->len; i++ ){
-	cattrs[i] = STR2CSTR(RARRAY(attrs)->ptr[i]);
-      };
-      cattrs[RARRAY(attrs)->len] = NULL;
-    }
-    else{
-      cattrs = NULL;
-    };
-    tv.tv_sec = 0;
-    tv.tv_usec = 0;
-    break;
   case 4:
-    cbase = STR2CSTR(base);
-    cscope = NUM2INT(scope);
-    cfilter = STR2CSTR(filter);
-    cattrsonly = 0; /* default value */
     if( TYPE(attrs) == T_ARRAY ){
-      cattrs = ALLOCA_N(char*,(RARRAY(attrs)->len + 1));
-      for( i=0; i < RARRAY(attrs)->len; i++ ){
-	cattrs[i] = STR2CSTR(RARRAY(attrs)->ptr[i]);
-      };
-      cattrs[RARRAY(attrs)->len] = NULL;
+      if( RARRAY(attrs)->len == 0 ){
+	cattrs = NULL;
+      }
+      else{
+	cattrs = ALLOCA_N(char*,(RARRAY(attrs)->len + 1));
+	for( i=0; i < RARRAY(attrs)->len; i++ ){
+	  cattrs[i] = StringValueCStr(RARRAY(attrs)->ptr[i]);
+	};
+	cattrs[RARRAY(attrs)->len] = NULL;
+      }
     }
     else{
       cattrs = NULL;
     };
-    tv.tv_sec = 0;  /* indicates the search method can't use timeout function */
-    tv.tv_usec = 0;
-    break;
   case 3:
-    cbase = STR2CSTR(base);
+    cbase = StringValueCStr(base);
     cscope = NUM2INT(scope);
-    cfilter = STR2CSTR(filter);
-    cattrsonly = 0; /* default value */
-    cattrs = NULL;  /* default value */
-    tv.tv_sec = 0;  /* indicates the search method can't use timeout function */
-    tv.tv_usec = 0;
+    cfilter = StringValueCStr(filter);
     break;
   default:
     rb_bug("rb_ldap_conn_search_s");
@@ -702,12 +642,212 @@ rb_ldap_conn_search_i(int argc, VALUE argv[], VALUE self,
   (*cmsg) = NULL;
   if( tv.tv_sec == 0 && tv.tv_usec == 0 ){
     (*ldapdata)->err = ldap_search_s(cldap, cbase, cscope, cfilter,
-				  cattrs, cattrsonly, cmsg);
+				     cattrs, cattrsonly, cmsg);
   }
   else{
     (*ldapdata)->err = ldap_search_st(cldap, cbase, cscope, cfilter,
 				      cattrs, cattrsonly, &tv, cmsg);
+  }
+  if( !(cmsg && (*cmsg)) ){
+      rb_raise(rb_eRuntimeError, "no message");
+  }
+  Check_LDAP_Result((*ldapdata)->err);
+
+#ifdef HAVE_LDAP_SORT_ENTRIES
+  if( rb_ldap_sort_obj != Qnil ){
+    ldap_sort_entries((*ldapdata)->ldap, cmsg,
+		      sort_attr, rb_ldap_internal_strcmp);
   };
+#endif
+  rb_ldap_sort_obj = Qnil;
+
+  return (*ldapdata)->err;
+}
+
+static VALUE
+rb_ldap_conn_search_b(VALUE rdata)
+{
+    void **data = (void**)rdata;
+    LDAP *cldap = (LDAP*)data[0];
+    LDAPMessage *cmsg = (LDAPMessage*)data[1];
+    LDAPMessage *e;
+    VALUE m;
+
+    for( e = ldap_first_entry(cldap, cmsg); e != NULL; e = ldap_next_entry(cldap, e) ){
+	m = rb_ldap_entry_new(cldap, e);
+	rb_ensure(rb_yield, m, rb_ldap_conn_invalidate_entry, m);
+    }
+    return Qnil;
+}
+
+static VALUE
+rb_ldap_msgfree(VALUE data)
+{
+    LDAPMessage *cmsg = (LDAPMessage*)data;
+    ldap_msgfree(cmsg);
+    return Qnil;
+}
+
+VALUE
+rb_ldap_conn_search_s(int argc, VALUE argv[], VALUE self)
+{
+    RB_LDAP_DATA *ldapdata;
+    LDAPMessage *cmsg;
+    LDAP *cldap;
+    LDAPMessage *e;
+
+    rb_ldap_conn_search_i(argc, argv, self, &ldapdata, &cmsg);
+    cldap = ldapdata->ldap;
+
+    if( ldapdata->err == LDAP_SUCCESS ){
+	void *pass_data[] = {(void*)cldap, (void*)cmsg};
+	rb_ensure(rb_ldap_conn_search_b, (VALUE)pass_data,
+		  rb_ldap_msgfree, (VALUE)cmsg);
+    };
+
+    return self;
+}
+
+static VALUE
+rb_ldap_conn_search2_b(VALUE rdata)
+{
+    void **data = (void*)rdata;
+    LDAP *cldap = (LDAP*)data[0];
+    LDAPMessage *cmsg = (LDAPMessage*)data[1];
+    VALUE ary = (VALUE)data[2];
+    LDAPMessage *e;
+    VALUE m;
+    VALUE hash;
+
+    for( e = ldap_first_entry(cldap, cmsg); e != NULL; e = ldap_next_entry(cldap, e) ){
+	m = rb_ldap_entry_new(cldap, e);
+	hash  = rb_ldap_entry_to_hash(m);
+	rb_ary_push(ary, hash);
+	if( rb_block_given_p() ){
+	    rb_ensure(rb_yield, hash, rb_ldap_conn_invalidate_entry, m);
+	}
+    }
+    return Qnil;
+}
+
+VALUE
+rb_ldap_conn_search2_s(int argc, VALUE argv[], VALUE self)
+{
+  RB_LDAP_DATA *ldapdata;
+  LDAPMessage *cmsg;
+  LDAPMessage *e;
+  LDAP *cldap;
+  VALUE ary = Qnil;
+
+  rb_ldap_conn_search_i(argc, argv, self, &ldapdata, &cmsg);
+  cldap = ldapdata->ldap;
+
+  ary = rb_ary_new();
+  if( ldapdata->err == LDAP_SUCCESS ){
+      void *pass_data[] = {(void*)cldap, (void*)cmsg, (void*)ary};
+      rb_ensure(rb_ldap_conn_search2_b, (VALUE)pass_data,
+		rb_ldap_msgfree, (VALUE)cmsg);
+  }
+
+  if( rb_block_given_p() ){
+    return self;
+  }
+  else{
+    return ary;
+  }
+}
+
+#if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_SEARCH_EXT_S)
+static int
+rb_ldap_conn_search_ext_i(int argc, VALUE argv[], VALUE self,
+			  RB_LDAP_DATA **ldapdata, LDAPMessage **cmsg)
+{
+  VALUE base, scope, filter, attrs, attrsonly;
+  VALUE serverctrls, clientctrls, sec, usec, limit, s_attr, s_proc;
+
+  RB_LDAPENTRY_DATA *edata;
+  LDAP *cldap;
+  char *cbase;
+  int cscope;
+  int climit;
+  char *cfilter;
+  char **cattrs;
+  char *sort_attr;
+  int cattrsonly;
+  VALUE msg;
+  int i,j;
+  struct timeval tv;
+  LDAPControl **sctrls, **cctrls;
+
+  GET_LDAP_DATA(self, (*ldapdata));
+  cldap = (*ldapdata)->ldap;
+
+  cattrs = NULL;
+  cattrsonly = 0;
+  cctrls = NULL;
+  sctrls = NULL;
+  tv.tv_sec = 0;
+  tv.tv_usec = 0;
+  sort_attr = NULL;
+  rb_ldap_sort_obj = Qnil;
+  climit = 0;
+
+  switch( rb_scan_args(argc,argv,"39",
+		       &base,&scope,&filter,&attrs,&attrsonly,
+		       &serverctrls,&clientctrls,&sec,&usec,&limit,
+		       &s_attr,&s_proc) ){
+  case 12:
+    rb_ldap_sort_obj = s_proc; /* Ruby's GC never start in a C function */
+  case 11:
+    if( rb_ldap_sort_obj == Qnil ){
+      rb_ldap_sort_obj = Qtrue;
+    }
+    sort_attr = StringValueCStr(s_attr);
+  case 10:
+    climit = NUM2INT(limit);
+  case 9:
+    tv.tv_usec = NUM2INT(usec);
+  case 8:
+    tv.tv_sec = NUM2INT(sec);
+  case 7:
+    cctrls = rb_ldap_get_controls(clientctrls);
+  case 6:
+    sctrls = rb_ldap_get_controls(serverctrls);
+  case 5:
+    cattrsonly = (attrsonly==Qtrue) ? 1 : 0;
+  case 4:
+    if( TYPE(attrs) == T_ARRAY ){
+      cattrs = ALLOCA_N(char*,(RARRAY(attrs)->len + 1));
+      for( i=0; i < RARRAY(attrs)->len; i++ ){
+	cattrs[i] = StringValueCStr(RARRAY(attrs)->ptr[i]);
+      };
+      cattrs[RARRAY(attrs)->len] = NULL;
+    }
+    else{
+      cattrs = NULL;
+    };
+  case 3:
+    cbase = StringValueCStr(base);
+    cscope = NUM2INT(scope);
+    cfilter = StringValueCStr(filter);
+    break;
+  default:
+    rb_bug("rb_ldap_conn_search_s");
+  };
+
+  (*cmsg) = NULL;
+  if( tv.tv_sec == 0 && tv.tv_usec == 0 ){
+    (*ldapdata)->err = ldap_search_ext_s(cldap, cbase, cscope, cfilter,
+					 cattrs, cattrsonly,
+					 sctrls, cctrls,
+					 NULL, climit, cmsg);
+  }
+  else{
+    (*ldapdata)->err = ldap_search_ext_s(cldap, cbase, cscope, cfilter,
+					 cattrs, cattrsonly,
+					 sctrls, cctrls,
+					 &tv, climit, cmsg);
+  }
   Check_LDAP_Result((*ldapdata)->err);
 
 #ifdef HAVE_LDAP_SORT_ENTRIES
@@ -722,67 +862,53 @@ rb_ldap_conn_search_i(int argc, VALUE argv[], VALUE self,
 };
 
 VALUE
-rb_ldap_conn_search_s(int argc, VALUE argv[], VALUE self)
+rb_ldap_conn_search_ext_s(int argc, VALUE argv[], VALUE self)
 {
   RB_LDAP_DATA *ldapdata;
   LDAPMessage *cmsg;
   LDAP *cldap;
   LDAPMessage *e;
 
-  rb_ldap_conn_search_i(argc, argv, self, &ldapdata, &cmsg);
+  rb_ldap_conn_search_ext_i(argc, argv, self, &ldapdata, &cmsg);
   cldap = ldapdata->ldap;
 
   if( ldapdata->err == LDAP_SUCCESS ){
-    for( e = ldap_first_entry(cldap, cmsg);
-	 e != NULL;
-	 e = ldap_next_entry(cldap, e) ){
-      VALUE m;
-      m = rb_ldap_entry_new(cldap, e);
-      rb_ensure(rb_yield, m, rb_ldap_conn_invalidate_entry, m);
-    }
-    ldap_msgfree(cmsg);
+      void *pass_data[] = {(void*)cldap, (void*)cmsg};
+      rb_ensure(rb_ldap_conn_search_b, (VALUE)pass_data,
+		rb_ldap_msgfree, (VALUE)cmsg);
   };
 
   return self;
 };
 
 VALUE
-rb_ldap_conn_search2_s(int argc, VALUE argv[], VALUE self)
+rb_ldap_conn_search_ext2_s(int argc, VALUE argv[], VALUE self)
 {
   RB_LDAP_DATA *ldapdata;
   LDAPMessage *cmsg;
-  LDAP *cldap;
-  VALUE ary, entry, hash;
   LDAPMessage *e;
-  VALUE result;
-  int has_block = rb_block_given_p();
+  LDAP *cldap;
+  VALUE ary;
 
-  rb_ldap_conn_search_i(argc, argv, self, &ldapdata, &cmsg);
+  rb_ldap_conn_search_ext_i(argc, argv, self, &ldapdata, &cmsg);
   cldap = ldapdata->ldap;
 
   ary = rb_ary_new();
-  result = ary;
   if( ldapdata->err == LDAP_SUCCESS ){
-    for( e = ldap_first_entry(cldap, cmsg);
-	 e != NULL;
-	 e = ldap_next_entry(cldap, e) ){
-      entry = rb_ldap_entry_new(cldap, e);
-      hash  = rb_funcall(entry, rb_intern("to_hash"), 0);
-      if( has_block ){
-	rb_yield(hash);
-      };
-      rb_ary_push(ary, hash);
-    }
-    ldap_msgfree(cmsg);
-  };
+      void *pass_data[] = {(void*)cldap, (void*)cmsg, (void*)ary};
+      rb_ensure(rb_ldap_conn_search2_b, (VALUE)pass_data,
+		rb_ldap_msgfree, (VALUE)cmsg);
+  }
 
-  if( has_block ){
+  if( rb_block_given_p() ){
     return self;
   }
   else{
-    return result;
-  };
-};
+    return ary;
+  }
+}
+#endif
+
 
 VALUE
 rb_ldap_conn_add_s(VALUE self, VALUE dn, VALUE attrs)
@@ -803,8 +929,8 @@ rb_ldap_conn_add_s(VALUE self, VALUE dn, VALUE attrs)
   };
 
   GET_LDAP_DATA(self, ldapdata);
-  c_dn = STR2CSTR(dn);
-  c_attrs = (LDAPMod**)malloc(sizeof(LDAPMod*) * (RARRAY(attrs)->len + 1));
+  c_dn = StringValueCStr(dn);
+  c_attrs = ALLOCA_N(LDAPMod*, (RARRAY(attrs)->len + 1));
 
   for ( i=0; i < RARRAY(attrs)->len; i++ ){
     VALUE mod = RARRAY(attrs)->ptr[i];
@@ -821,6 +947,49 @@ rb_ldap_conn_add_s(VALUE self, VALUE dn, VALUE attrs)
   return self;
 };
 
+#if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_ADD_EXT_S)
+VALUE
+rb_ldap_conn_add_ext_s(VALUE self, VALUE dn, VALUE attrs,
+		       VALUE serverctrls, VALUE clientctrls)
+{
+  RB_LDAP_DATA *ldapdata;
+  char *c_dn;
+  LDAPMod **c_attrs;
+  int i;
+  LDAPControl **sctrls, **cctrls;
+
+  switch( TYPE(attrs) ){
+  case T_HASH:
+    attrs = rb_ldap_hash2mods(rb_mLDAP, INT2NUM(LDAP_MOD_ADD), attrs);
+    break;
+  case T_ARRAY:
+    break;
+  default:
+    rb_raise(rb_eTypeError, "must be a hash or array");
+  };
+
+  GET_LDAP_DATA(self, ldapdata);
+  c_dn = StringValueCStr(dn);
+  c_attrs = ALLOCA_N(LDAPMod*, (RARRAY(attrs)->len + 1));
+  sctrls = rb_ldap_get_controls(serverctrls);
+  cctrls = rb_ldap_get_controls(clientctrls);
+
+  for ( i=0; i < RARRAY(attrs)->len; i++ ){
+    VALUE mod = RARRAY(attrs)->ptr[i];
+    RB_LDAPMOD_DATA *moddata;
+    Check_Kind(mod, rb_cLDAP_Mod);
+    GET_LDAPMOD_DATA(mod, moddata);
+    c_attrs[i] = moddata->mod;
+  };
+  c_attrs[i] = NULL;
+
+  ldapdata->err = ldap_add_ext_s(ldapdata->ldap, c_dn, c_attrs, sctrls, cctrls);
+  Check_LDAP_Result(ldapdata->err);
+
+  return self;
+}
+#endif
+
 VALUE
 rb_ldap_conn_modify_s(VALUE self, VALUE dn, VALUE attrs)
 {
@@ -840,8 +1009,8 @@ rb_ldap_conn_modify_s(VALUE self, VALUE dn, VALUE attrs)
   };
 
   GET_LDAP_DATA(self, ldapdata);
-  c_dn = STR2CSTR(dn);
-  c_attrs = (LDAPMod**)malloc(sizeof(LDAPMod*) * (RARRAY(attrs)->len + 1));
+  c_dn = StringValueCStr(dn);
+  c_attrs = ALLOC_N(LDAPMod*, RARRAY(attrs)->len + 1);
 
   for ( i=0; i < RARRAY(attrs)->len; i++ ){
     VALUE mod = RARRAY(attrs)->ptr[i];
@@ -858,6 +1027,49 @@ rb_ldap_conn_modify_s(VALUE self, VALUE dn, VALUE attrs)
   return self;
 };
 
+#if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_MODIFY_EXT_S)
+VALUE
+rb_ldap_conn_modify_ext_s(VALUE self, VALUE dn, VALUE attrs,
+			  VALUE serverctrls, VALUE clientctrls)
+{
+  RB_LDAP_DATA *ldapdata;
+  char *c_dn;
+  LDAPMod **c_attrs;
+  int i;
+  LDAPControl **sctrls, **cctrls;
+
+  switch( TYPE(attrs) ){
+  case T_HASH:
+    attrs = rb_ldap_hash2mods(rb_mLDAP, INT2NUM(LDAP_MOD_REPLACE), attrs);
+    break;
+  case T_ARRAY:
+    break;
+  default:
+    rb_raise(rb_eTypeError, "must be a hash or array");
+  };
+
+  GET_LDAP_DATA(self, ldapdata);
+  c_dn = StringValueCStr(dn);
+  c_attrs = ALLOC_N(LDAPMod*, RARRAY(attrs)->len + 1);
+  sctrls = rb_ldap_get_controls(serverctrls);
+  cctrls = rb_ldap_get_controls(clientctrls);
+
+  for ( i=0; i < RARRAY(attrs)->len; i++ ){
+    VALUE mod = RARRAY(attrs)->ptr[i];
+    RB_LDAPMOD_DATA *moddata;
+    Check_Kind(mod, rb_cLDAP_Mod);
+    GET_LDAPMOD_DATA(mod, moddata);
+    c_attrs[i] = moddata->mod;
+  };
+  c_attrs[i] = NULL;
+
+  ldapdata->err = ldap_modify_ext_s(ldapdata->ldap, c_dn, c_attrs, sctrls, cctrls);
+  Check_LDAP_Result(ldapdata->err);
+
+  return self;
+}
+#endif
+
 VALUE
 rb_ldap_conn_modrdn_s(VALUE self, VALUE dn, VALUE newrdn, VALUE delete_p)
 {
@@ -867,8 +1079,8 @@ rb_ldap_conn_modrdn_s(VALUE self, VALUE dn, VALUE newrdn, VALUE delete_p)
   int c_delete_p;
 
   GET_LDAP_DATA(self, ldapdata);
-  c_dn = STR2CSTR(dn);
-  c_newrdn = STR2CSTR(newrdn);
+  c_dn = StringValueCStr(dn);
+  c_newrdn = StringValueCStr(newrdn);
   c_delete_p = (delete_p == Qtrue) ? 1 : 0;
 
   ldapdata->err = ldap_modrdn2_s(ldapdata->ldap, c_dn, c_newrdn, c_delete_p);
@@ -884,7 +1096,7 @@ rb_ldap_conn_delete_s(VALUE self, VALUE dn)
   char *c_dn;
 
   GET_LDAP_DATA(self, ldapdata);
-  c_dn = STR2CSTR(dn);
+  c_dn = StringValueCStr(dn);
 
   ldapdata->err = ldap_delete_s(ldapdata->ldap, c_dn);
   Check_LDAP_Result(ldapdata->err);
@@ -892,27 +1104,72 @@ rb_ldap_conn_delete_s(VALUE self, VALUE dn)
   return self;
 };
 
+#if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_DELETE_EXT_S)
+VALUE
+rb_ldap_conn_delete_ext_s(VALUE self, VALUE dn,
+			  VALUE serverctrls, VALUE clientctrls)
+{
+  RB_LDAP_DATA *ldapdata;
+  char *c_dn;
+  LDAPControl **sctrls, **cctrls;
+
+  GET_LDAP_DATA(self, ldapdata);
+  c_dn = StringValueCStr(dn);
+  sctrls = rb_ldap_get_controls(serverctrls);
+  cctrls = rb_ldap_get_controls(clientctrls);
+
+  ldapdata->err = ldap_delete_ext_s(ldapdata->ldap, c_dn, sctrls, cctrls);
+  Check_LDAP_Result(ldapdata->err);
+
+  return self;
+}
+#endif
+
+#if defined(HAVE_LDAP_COMPARE_S)
 VALUE
 rb_ldap_conn_compare_s(VALUE self, VALUE dn, VALUE attr, VALUE val)
 {
-#if defined(LDAP_COMPARE_S)
   RB_LDAP_DATA *ldapdata;
   char *c_dn, *c_attr, *c_val;
 
   GET_LDAP_DATA(self, ldapdata);
-  c_dn = STR2CSTR(dn);
-  c_attr = STR2CSTR(attr);
-  c_val = STR2CSTR(val);
+  c_dn = StringValueCStr(dn);
+  c_attr = StringValueCStr(attr);
+  c_val = StringValueCStr(val);
 
   ldapdata->err = ldap_compare_s(ldapdata->ldap, c_dn, c_attr, c_val);
   Check_LDAP_Result(ldapdata->err);
 
   return self;
-#else
-  rb_notimplement();
+}
+#endif
+
+#if defined(HAVE_LDAPCONTROL) && defined(HAVE_LDAP_COMPARE_EXT_S)
+VALUE
+rb_ldap_conn_compare_ext_s(VALUE self, VALUE dn, VALUE attr, VALUE val,
+			   VALUE serverctrls, VALUE clientctrls)
+{
+  RB_LDAP_DATA *ldapdata;
+  char *c_dn, *c_attr;
+  struct berval bval;
+  LDAPControl **sctrls, **cctrls;
+
+  GET_LDAP_DATA(self, ldapdata);
+  c_dn = StringValueCStr(dn);
+  c_attr = StringValueCStr(attr);
+  bval.bv_val = StringValueCStr(val);
+  bval.bv_len = RSTRING(val)->len;
+  sctrls = rb_ldap_get_controls(serverctrls);
+  cctrls = rb_ldap_get_controls(clientctrls);
+
+  ldapdata->err = ldap_compare_ext_s(ldapdata->ldap,
+				     c_dn, c_attr, &bval,
+				     sctrls, cctrls);
+  Check_LDAP_Result(ldapdata->err);
+
   return self;
+}
 #endif
-};
 
 VALUE
 rb_ldap_conn_err(VALUE self)
@@ -928,8 +1185,15 @@ Init_ldap_conn()
 {
   rb_ldap_sort_obj = Qnil;
 
-  rb_cLDAP_Conn = rb_define_class_under(rb_mLDAP,"Conn",rb_cObject);
-  rb_define_singleton_method(rb_cLDAP_Conn,"new",rb_ldap_conn_s_new,-1);
+  rb_cLDAP_Conn = rb_define_class_under(rb_mLDAP,"Conn",rb_cData);
+#if RUBY_VERSION_CODE < 170
+  rb_define_singleton_method(rb_cLDAP_Conn,"new",rb_ldap_class_new,-1);
+#endif
+#if RUBY_VERSION_CODE >= 173
+  rb_define_alloc_func(rb_cLDAP_Conn,rb_ldap_conn_s_allocate);
+#else
+  rb_define_singleton_method(rb_cLDAP_Conn,"allocate",rb_ldap_conn_s_allocate,0);
+#endif
   rb_define_singleton_method(rb_cLDAP_Conn,"open",rb_ldap_conn_s_open,-1);
   rb_ldap_conn_define_method("initialize", rb_ldap_conn_initialize,-1);
   rb_ldap_conn_define_method("start_tls", rb_ldap_conn_start_tls_s, -1);
@@ -944,9 +1208,30 @@ Init_ldap_conn()
   rb_ldap_conn_define_method("modify", rb_ldap_conn_modify_s, 2);
   rb_ldap_conn_define_method("modrdn", rb_ldap_conn_modrdn_s, 3);
   rb_ldap_conn_define_method("delete", rb_ldap_conn_delete_s, 1);
+#if defined(HAVE_LDAP_COMPARE_S)
   rb_ldap_conn_define_method("compare", rb_ldap_conn_compare_s, 3);
+#endif
   rb_ldap_conn_define_method("perror", rb_ldap_conn_perror, 1);
   rb_ldap_conn_define_method("err2string", rb_ldap_conn_err2string, 1);
   rb_ldap_conn_define_method("result2error", rb_ldap_conn_result2error, 1);
   rb_ldap_conn_define_method("err", rb_ldap_conn_err, 0);
+
+#if defined(HAVE_LDAPCONTROL)
+#if defined(HAVE_LDAP_SEARCH_EXT_S)
+  rb_ldap_conn_define_method("search_ext", rb_ldap_conn_search_ext_s, -1);
+  rb_ldap_conn_define_method("search_ext2", rb_ldap_conn_search_ext2_s, -1);
+#endif
+#if defined(HAVE_LDAP_ADD_EXT_S)
+  rb_ldap_conn_define_method("add_ext", rb_ldap_conn_add_ext_s, 4);
+#endif
+#if defined(HAVE_LDAP_MODIFY_EXT_S)
+  rb_ldap_conn_define_method("modify_ext", rb_ldap_conn_modify_ext_s, 4);
+#endif
+#if defined(HAVE_LDAP_DELETE_EXT_S)
+  rb_ldap_conn_define_method("delete_ext", rb_ldap_conn_delete_ext_s, 3);
+#endif
+#if defined(HAVE_LDAP_COMPARE_EXT_S)
+  rb_ldap_conn_define_method("compare_ext", rb_ldap_conn_compare_ext_s, 5);
+#endif
+#endif
 };
