@@ -1,17 +1,17 @@
 # -*- ruby -*-
 # This file is a part of test scripts of LDAP extension module.
 
-require "test/conf"
-require "ldap"
+$test = File.dirname($0)
+require "#{$test}/conf"
+require "./ldap"
 
 conn = LDAP::Conn.new($HOST, $PORT)
 conn.bind('cn=root, dc=localhost, dc=localdomain','secret'){
   conn.perror("bind")
   entry1 = [
-    LDAP.mod(LDAP::LDAP_MOD_ADD,
-	     'objectclass', ['top', 'organization']),
-    LDAP.mod(LDAP::LDAP_MOD_ADD,
-	     'o', ['Japan Advanced Institute of Science and Technology']),
+    LDAP.mod(LDAP::LDAP_MOD_ADD, 'objectclass', ['top', 'domain']),
+    LDAP.mod(LDAP::LDAP_MOD_ADD, 'o', ['TTSKY.NET']),
+    LDAP.mod(LDAP::LDAP_MOD_ADD, 'dc', ['localhost']),
   ]
 
   entry2 = [
